import config.base

import os
import re

class Configure(config.base.Configure):
  def __init__(self, framework, libraries = []):
    config.base.Configure.__init__(self, framework)
    self.headerPrefix  = ''
    self.substPrefix   = ''
    self.libraries     = libraries
    self.rpathSkipDirs = [] # do not generate RPATH for dirs in this list; useful when compiling with stub libraries (.so) that do not have corresponding runtime library (.so.1) at this location. Check cuda.py for usage.
    self.sysDirs       = ['/usr/lib','/lib','/usr/lib64','/lib64'] # skip conversion from full path to link line argument format for libraries in these dirs. For ex: some compilers internally use /usr/lib/libm.so that should not be converted to '-L/usr/lib -lm'
    return

  def setupDependencies(self, framework):
    config.base.Configure.setupDependencies(self, framework)
    self.setCompilers = framework.require('config.setCompilers', self)
    self.compilers    = framework.require('config.compilers',    self)
    self.headers      = framework.require('config.headers',      self)
    self.types        = framework.require('config.types',        self)
    return

  def getLibArgumentList(self, library, with_rpath=True):
    '''Return the proper link line argument for the given filename library as a list of options
      - If the path is empty, return it unchanged
      - If starts with - then return unchanged
      - If the path ends in ".lib" return it unchanged
      - If the path is absolute and the filename is "lib"<name>, return -L<dir> -l<name> (optionally including rpath flag)
      - If the filename is "lib"<name>, return -l<name>
      - If the path ends in ".so" or ".dylib" return it unchanged
      - If the path ends in ".o" return it unchanged
      - If the path is absolute, return it unchanged
      - Otherwise return -l<library>'''
    if not library:
      return []
    if library.startswith('${CC_LINKER_SLFLAG}'):
      return [library] if with_rpath else []
    if library.startswith('${FC_LINKER_SLFLAG}'):
      return [library] if with_rpath else []
    flagName  = self.language[-1]+'SharedLinkerFlag'
    flagSubst = self.language[-1].upper()+'_LINKER_SLFLAG'
    rpathFlag = ''
    if hasattr(self.setCompilers, flagName) and not getattr(self.setCompilers, flagName) is None:
      rpathFlag = getattr(self.setCompilers, flagName)
    elif flagSubst in self.argDB:
      rpathFlag = self.argDB[flagSubst]
    if library.startswith('-L'):
      dirname = library[2:]
      if not dirname.startswith('$') and not os.path.isdir(dirname): self.logPrint('Warning! getLibArgumentList(): could not locate dir '+ dirname)
      if dirname in self.sysDirs:
          return []
      elif with_rpath and rpathFlag and not dirname in self.rpathSkipDirs:
        return [rpathFlag+dirname,library]
      else:
        return [library]
    if library.lstrip()[0] == '-':
      return [library]
    if len(library) > 3 and library[-4:] == '.lib':
      return [library.replace('\\ ',' ').replace(' ', '\\ ').replace('\\(','(').replace('(', '\\(').replace('\\)',')').replace(')', '\\)')]
    if os.path.basename(library).startswith('lib'):
      name = self.getLibName(library)
      if ((len(library) > 2 and library[1] == ':') or os.path.isabs(library)):
        dirname   = os.path.dirname(library).replace('\\ ',' ').replace(' ', '\\ ').replace('\\(','(').replace('(', '\\(').replace('\\)',')').replace(')', '\\)')
        if dirname in self.sysDirs:
          return [library]
        if with_rpath and not dirname in self.rpathSkipDirs:
          if hasattr(self.setCompilers, flagName) and not getattr(self.setCompilers, flagName) is None:
            import pathlib
            if pathlib.Path(library).suffix[1:].isnumeric(): # libfoo.so.1.0
              return [getattr(self.setCompilers, flagName)+dirname,library]
            else:
              return [getattr(self.setCompilers, flagName)+dirname,'-L'+dirname,'-l'+name]
          if flagSubst in self.argDB:
            return [self.argDB[flagSubst]+dirname,'-L'+dirname,'-l'+name]
        return ['-L'+dirname,'-l'+name]
      else:
        return ['-l'+name]
    if os.path.splitext(library)[1] == '.so' or os.path.splitext(library)[1] == '.o' or os.path.splitext(library)[1] == '.dylib':
      return [library]
    if os.path.isabs(library):
      return [library]
    return ['-l'+library]

  def getLibArgument(self, library):
    '''Same as getLibArgumentList - except it returns a string instead of list.'''
    return  ' '.join(self.getLibArgumentList(library))

  def addRpathSkipDir(self, dirname):
    '''Do not generate RPATH for this dir in getLibArgumentList.'''
    if dirname not in self.rpathSkipDirs: self.rpathSkipDirs.append(dirname)

  def addSysDir(self, dirname):
    '''Add the dir to sysDirs[]'''
    if dirname not in self.sysDirs: self.sysDirs.append(dirname)

  def getLibName(library):
    if os.path.basename(library).startswith('lib'):
      return os.path.splitext(os.path.basename(library))[0][3:]
    return library
  getLibName = staticmethod(getLibName)

  def getDefineName(self, library):
    return 'HAVE_LIB'+self.getLibName(library).upper().replace('-','_').replace('=','_').replace('+','_').replace('.', '_').replace('/','_')

  def getDefineNameFunc(self, funcName):
    return 'HAVE_'+ funcName.upper()

  def haveLib(self, library):
    return self.getDefineName(library) in self.defines

  def add(self, libName, funcs, libDir = None, otherLibs = [], prototype = '', call = '', fortranMangle = 0):
    '''Checks that the library "libName" contains "funcs", and if it does defines HAVE_LIB"libName AND adds it to $LIBS"
       - libDir may be a list of directories
       - libName may be a list of library names'''
    if not isinstance(libName, list): libName = [libName]
    if self.check(libName, funcs, libDir, otherLibs, prototype, call, fortranMangle):
      self.logPrint('Adding '+str(libName)+' to LIBS')
      # Note: this MUST be setCompilers since it can happen before dispatch names is made
      self.setCompilers.LIBS = self.toString(libName)+' '+self.setCompilers.LIBS
      return 1
    return 0

  def toString(self,libs):
    '''Converts a list of libraries to a string suitable for a linker'''
    newlibs = []
    frame = 0
    for lib in libs:
      if frame:
        newlibs += [lib]
        frame   = 0
      elif lib == '-framework':
        newlibs += [lib]
        frame = 1
      else:
        newlibs += self.getLibArgumentList(lib)
    return ' '.join(newlibs)

  def toStringNoDupes(self,libs,with_rpath=True):
    '''Converts a list of libraries to a string suitable for a linker, removes duplicates'''
    '''Moves the flags that can be moved to the beginning of the string but always leaves the libraries and other items that must remain in the same order'''
    newlibs = []
    frame = 0
    for lib in libs:
      if frame:
        newlibs += [lib]
        frame   = 0
      elif lib == '-framework':
        newlibs += [lib]
        frame = 1
      else:
        newlibs += self.getLibArgumentList(lib, with_rpath)
    libs = newlibs
    newldflags = []
    newlibs = []
    frame = 0
    dupflags = ['-L']
    flagName  = self.language[-1]+'SharedLinkerFlag'
    if hasattr(self.setCompilers, flagName) and not getattr(self.setCompilers, flagName) is None:
      dupflags.append(getattr(self.setCompilers, flagName))
    for j in libs:
      # remove duplicate -L, -Wl,-rpath options - and only consecutive -l options
      if j in newldflags and any([j.startswith(flg) for flg in dupflags]): continue
      if newlibs and j == newlibs[-1]: continue
      if list(filter(j.startswith,['-l'])) or list(filter(j.endswith,['.lib','.a','.so','.o'])) or j in ['-Wl,-Bstatic','-Wl,-Bdynamic','-Wl,--start-group','-Wl,--end-group']:
        newlibs.append(j)
      else:
        newldflags.append(j)
    liblist = ' '.join(newldflags + newlibs)
    return liblist

  def getShortLibName(self,lib):
    '''returns the short name for the library. Valid names are foo -lfoo or libfoo.[a,so,lib]'''
    if lib.startswith('-l'):
      libname = lib[2:]
      return libname
    if lib.startswith('-'): # must be some compiler options - not a library
      return ''
    if lib.endswith('.a') or lib.endswith('.so') or lib.endswith('.lib'):
      libname = os.path.splitext(os.path.basename(lib))[0]
      if lib.startswith('lib'): libname = libname[3:]
      return libname
    # no match - assuming the given name is already in short notation
    return lib

  def check(self, libName, funcs, libDir = None, otherLibs = [], prototype = '', call = '', fortranMangle = 0, cxxMangle = 0, cxxLink = 0, functionDefine = 0, examineOutput=lambda ret,out,err:None):
    '''Checks that the library "libName" contains "funcs", and if it does defines HAVE_LIB"libName"
       - libDir may be a list of directories
       - libName may be a list of library names'''
    if not isinstance(funcs,list): funcs = [funcs]
    if not isinstance(libName, list): libName = [libName]
    def genPreamble(f, funcName):
      # Construct prototype
      if self.language[-1] == 'FC':
        return ''
      if prototype:
        if isinstance(prototype, str):
          pre = prototype
        else:
          pre = prototype[f]
      else:
        # We use char because int might match the return type of a gcc2 builtin and its argument prototype would still apply.
        pre = 'char '+funcName+'(void);'
      # Capture the function call in a static function so that any local variables are isolated from
      # calls to other library functions.
      return pre + '\nstatic void _check_%s(void) { %s }' % (funcName, genCall(f, funcName, pre=True))
    def genCall(f, funcName, pre=False):
      if self.language[-1] != 'FC' and not pre:
        return '_check_' + funcName + '();'
      # Construct function call
      if call:
        if isinstance(call, str):
          body = call
        else:
          body = call[f]
      else:
        body = funcName+'()'
      if self.language[-1] != 'FC':
        body += ';'
      return body
    # Handle Fortran mangling
    if fortranMangle:
      funcs = list(map(self.compilers.mangleFortranFunction, funcs))
    if not funcs:
      self.logPrint('No functions to check for in library '+str(libName)+' '+str(otherLibs))
      return True
    self.logPrint('Checking for functions ['+' '.join(funcs)+'] in library '+str(libName)+' '+str(otherLibs))
    if self.language[-1] == 'FC':
      includes = ''
    else:
      includes = '/* Override any gcc2 internal prototype to avoid an error. */\n'
    # Handle C++ mangling
    if self.language[-1] == 'Cxx' and not cxxMangle:
      includes += '''
#ifdef __cplusplus
extern "C" {
#endif
'''
    includes += '\n'.join([genPreamble(f, fname) for f, fname in enumerate(funcs)])
    # Handle C++ mangling
    if self.language[-1] == 'Cxx' and not cxxMangle:
      includes += '''
#ifdef __cplusplus
}
#endif
'''
    body = '\n'.join([genCall(f, fname) for f, fname in enumerate(funcs)])
    # Setup link line
    oldLibs = self.setCompilers.LIBS
    if libDir:
      if not isinstance(libDir, list): libDir = [libDir]
      for dir in libDir:
        self.setCompilers.LIBS += ' -L'+dir
    # new libs may/will depend on system libs so list new libs first!
    # Matt, do not change this without talking to me
    if libName and otherLibs:
      self.setCompilers.LIBS = ' '+self.toString(libName+otherLibs) +' '+ self.setCompilers.LIBS
    elif otherLibs:
      self.setCompilers.LIBS = ' '+self.toString(otherLibs) +' '+ self.setCompilers.LIBS
    elif libName:
      self.setCompilers.LIBS = ' '+self.toString(libName) +' '+ self.setCompilers.LIBS
    if cxxMangle: compileLang = 'Cxx'
    else:         compileLang = self.language[-1]
    if cxxLink: linklang = 'Cxx'
    else: linklang = self.language[-1]
    self.pushLanguage(compileLang)

    found = 1
    if libName and libName[0].startswith('/'):
      dir = os.path.dirname(libName[0])
      lib = os.path.basename(libName[0])[:-1]
      self.logPrint('Checking directory of requested libraries:'+dir+' for first library:'+lib)
      found = 0
      try:
        files = os.listdir(dir)
      except:
        self.logPrint('Directory of requested libraries '+dir+' does not exist')
      else:
        self.logPrint('Files in directory:'+str(files))
        for i in files:
          if i.startswith(lib):
            found = 1
            break

    if found and self.checkLink(includes, body, linkLanguage=linklang, examineOutput=examineOutput):
      if hasattr(self.compilers, 'FC') and self.language[-1] == 'C':
        if self.compilers.checkCrossLink(includes+'\nvoid dummy(void) {'+body+'}\n',"     program main\n      print*,'testing'\n      stop\n      end\n",language1='C',language2='FC'):
          # define the symbol as found
          if functionDefine: [self.addDefine(self.getDefineNameFunc(fname), 1) for f, fname in enumerate(funcs)]
          # add to list of found libraries
          elif libName:
            for lib in libName:
              shortlib = self.getShortLibName(lib)
              if shortlib: self.addDefine(self.getDefineName(shortlib), 1)
        else:
          found = 0
    else:
      found = 0
    self.setCompilers.LIBS = oldLibs
    self.popLanguage()
    return found

  def checkClassify(self, libName, funcs, libDir=None, otherLibs=[], prototype='', call='', fortranMangle=0, cxxMangle=0, cxxLink=0):
    '''Recursive decompose to rapidly classify functions as found or missing'''
    import config
    def functional(funcs):
      named = config.NamedInStderr(funcs)
      if self.check(libName, funcs, libDir, otherLibs, prototype, call, fortranMangle, cxxMangle, cxxLink):
        return True
      else:
        return named.named
    found, missing = config.classify(funcs, functional)
    return found, missing

  def checkMath(self):
    '''Check for sin() in libm, the math library'''
    self.math = None
    funcs = ['sin', 'floor', 'log10', 'pow']
    prototypes = ['#include <stdio.h>\ndouble sin(double);',
                  '#include <stdio.h>\ndouble floor(double);',
                  '#include <stdio.h>\ndouble log10(double);',
                  '#include <stdio.h>\ndouble pow(double, double);']
    calls = ['double x,y; int s = scanf("%lf",&x); y = sin(x); printf("%f %d",y,s)',
             'double x,y; int s = scanf("%lf",&x); y = floor(x); printf("%f %d",y,s)',
             'double x,y; int s = scanf("%lf",&x); y = log10(x); printf("%f %d",y,s)',
             'double x,y; int s = scanf("%lf",&x); y = pow(x,x); printf("%f %d",y,s)']
    if self.check('', funcs, prototype = prototypes, call = calls):
      self.math = []
    elif self.check('m', funcs, prototype = prototypes, call = calls):
      self.math = ['libm.a']
    if self.math == None:
      raise RuntimeError('Cannot find basic math functions')
    self.logPrint('CheckMath: using math library '+str(self.math))
    return

  def checkMathErf(self):
    '''Check for erf() in libm, the math library'''
    if not self.math is None and self.check(self.math, ['erf'], prototype = ['#include <math.h>'], call = ['double (*checkErf)(double) = erf;double x = 0,y; y = (*checkErf)(x); (void)y']):
      self.logPrint('erf() found')
      self.addDefine('HAVE_ERF', 1)
    else:
      self.logPrint('erf() not found')
    return

  def checkMathTgamma(self):
    '''Check for tgamma() in libm, the math library'''
    if not self.math is None and self.check(self.math, ['tgamma'], prototype = ['#include <math.h>'], call = ['double (*checkTgamma)(double) = tgamma;double x = 0,y; y = (*checkTgamma)(x); (void)y']):
      self.logPrint('tgamma() found')
      self.addDefine('HAVE_TGAMMA', 1)
    else:
      self.logPrint('tgamma() not found')
    return

  def checkMathLgamma(self):
    '''Check for lgamma() in libm, the math library'''
    if not self.math is None and self.check(self.math, ['lgamma'], prototype = ['#include <math.h>\n#include <stdlib.h>'], call = ['double (*checkLgamma)(double) = lgamma;double x = 1,y; y = (*checkLgamma)(x);if (y != 0.) abort()']):
      self.logPrint('lgamma() found')
      self.addDefine('HAVE_LGAMMA', 1)
    elif not self.math is None and self.check(self.math, ['gamma'], prototype = ['#include <math.h>\n#include <stdlib.h>'], call = ['double (*checkLgamma)(double) = gamma;double x = 1,y; y = (*checkLgamma)(x);if (y != 0.) abort()']):
      self.logPrint('gamma() found')
      self.addDefine('HAVE_LGAMMA', 1)
      self.addDefine('HAVE_LGAMMA_IS_GAMMA', 1)
    else:
      self.logPrint('lgamma() and gamma() not found')
    return

  def checkMathFenv(self):
    '''Checks if <fenv.h> can be used with FE_DFL_ENV'''
    if not self.math is None and self.check(self.math, ['fesetenv'], prototype = ['#include <fenv.h>'], call = ['fesetenv(FE_DFL_ENV)']):
      self.addDefine('HAVE_FENV_H', 1)
    else:
      self.logPrint('<fenv.h> with FE_DFL_ENV not found')
    if not self.math is None and self.check(self.math, ['feclearexcept'], prototype = ['#include <fenv.h>'], call = ['feclearexcept(FE_INEXACT)']):
      self.addDefine('HAVE_FE_VALUES', 1)
    else:
      self.logPrint('<fenv.h> with FE_INEXACT not found')
    return

  def checkMathLog2(self):
    '''Check for log2() in libm, the math library'''
    if not self.math is None and self.check(self.math, ['log2'], prototype = ['#include <math.h>'], call = ['double (*checkLog2)(double) = log2; double x = 2.5, y = (*checkLog2)(x); (void)y']):
      self.logPrint('log2() found')
      self.addDefine('HAVE_LOG2', 1)
    else:
      self.logPrint('log2() not found')
    return

  def checkRealtime(self):
    '''Check for presence of clock_gettime() in realtime library (POSIX Realtime extensions)'''
    self.rt = None
    funcs = ['clock_gettime']
    prototypes = ['#include <time.h>']
    calls = ['struct timespec tp; clock_gettime(CLOCK_REALTIME,&tp)']
    if self.check('', funcs, prototype=prototypes, call=calls):
      self.logPrint('realtime functions are linked in by default')
      self.rt = []
    elif self.check('rt', funcs, prototype=prototypes, call=calls):
      self.logPrint('Using librt for the realtime library')
      self.rt = ['librt.a']
    else:
      self.logPrint('No realtime library found')
    return

  def checkDynamic(self):
    '''Check for the header and libraries necessary for dynamic library manipulation'''
    if 'with-dynamic-loading' in self.argDB and not self.argDB['with-dynamic-loading']: return
    self.check(['dl'], 'dlopen')
    self.headers.check('dlfcn.h')
    return

  def checkShared(self, includes, initFunction, checkFunction, finiFunction = None, checkLink = None, libraries = [], initArgs = '&argc, &argv', boolType = 'int', noCheckArg = 0, defaultArg = '', executor = None, timeout = 60):
    '''Determine whether a library is shared
       - initFunction(int *argc, char *argv[]) is called to initialize some static data
       - checkFunction(int *check) is called to verify that the static data wer set properly
       - finiFunction() is called to finalize the data, and may be omitted
       - checkLink may be given as ana alternative to the one in base.Configure'''
    isShared = 0
    if checkLink is None:
      checkLink = self.checkLink
      configObj = self
    else:
      if hasattr(checkLink, '__self__'):
        configObj = checkLink.__self__
      else:
        configObj = self

    # Fix these flags
    oldFlags = self.setCompilers.LIBS
    self.setCompilers.LIBS = ' '+self.toString(libraries)+' '+self.setCompilers.LIBS

    # Make a library which calls initFunction(), and returns checkFunction()
    lib1Name = os.path.join(self.tmpDir, 'lib1.'+self.setCompilers.sharedLibraryExt)
    if noCheckArg:
      checkCode = 'isInitialized = '+checkFunction+'();'
    else:
      checkCode = checkFunction+'(&isInitialized);'
    codeBegin = '''
#ifdef __cplusplus
extern "C"
#endif
int init(int argc,  char *argv[]) {
'''
    body      = '''
  %s isInitialized;

  %s(%s);
  %s
  return (int) isInitialized;
''' % (boolType, initFunction, initArgs, checkCode)
    codeEnd   = '\n}\n'
    if not checkLink(includes, body, cleanup = 0, codeBegin = codeBegin, codeEnd = codeEnd, shared = 1):
      if os.path.isfile(configObj.compilerObj): os.remove(configObj.compilerObj)
      self.setCompilers.LIBS = oldFlags
      raise RuntimeError('Could not complete shared library check')
    if os.path.isfile(configObj.compilerObj): os.remove(configObj.compilerObj)
    os.rename(configObj.linkerObj, lib1Name)

    # Make a library which calls checkFunction()
    lib2Name = os.path.join(self.tmpDir, 'lib2.'+self.setCompilers.sharedLibraryExt)
    codeBegin = '''
#ifdef __cplusplus
extern "C"
#endif
int checkInit(void) {
'''
    body      = '''
  %s isInitialized;

  %s
''' % (boolType, checkCode)
    if finiFunction:
      body += '  if (isInitialized) '+finiFunction+'();\n'
    body += '  return (int) isInitialized;\n'
    codeEnd   = '\n}\n'
    if not checkLink(includes, body, cleanup = 0, codeBegin = codeBegin, codeEnd = codeEnd, shared = 1):
      if os.path.isfile(configObj.compilerObj): os.remove(configObj.compilerObj)
      self.setCompilers.LIBS = oldFlags
      raise RuntimeError('Could not complete shared library check')
      return 0
    if os.path.isfile(configObj.compilerObj): os.remove(configObj.compilerObj)
    os.rename(configObj.linkerObj, lib2Name)

    self.setCompilers.LIBS = oldFlags

    # Make an executable that dynamically loads and calls both libraries
    #   If the check returns true in the second library, the static data was shared
    guard = self.headers.getDefineName('dlfcn.h')
    if self.headers.headerPrefix:
      guard = self.headers.headerPrefix+'_'+guard
    defaultIncludes = '''
#include <stdio.h>
#include <stdlib.h>
#ifdef %s
#include <dlfcn.h>
#endif
    ''' % guard
    body = '''
  int   argc    = 1;
  char *argv[2] = {(char *) "conftest", NULL};
  void *lib;
  int (*init)(int, char **);
  int (*checkInit)(void);

  lib = dlopen("'''+lib1Name+'''", RTLD_LAZY);
  if (!lib) {
    fprintf(stderr, "Could not open lib1.so: %s\\n", dlerror());
    exit(1);
  }
  init = (int (*)(int, char **)) dlsym(lib, "init");
  if (!init) {
    fprintf(stderr, "Could not find initialization function\\n");
    exit(1);
  }
  if (!(*init)(argc, argv)) {
    fprintf(stderr, "Could not initialize library\\n");
    exit(1);
  }
  lib = dlopen("'''+lib2Name+'''", RTLD_LAZY);
  if (!lib) {
    fprintf(stderr, "Could not open lib2.so: %s\\n", dlerror());
    exit(1);
  }
  checkInit = (int (*)(void)) dlsym(lib, "checkInit");
  if (!checkInit) {
    fprintf(stderr, "Could not find initialization check function\\n");
    exit(1);
  }
  if (!(*checkInit)()) {
    fprintf(stderr, "Did not link with shared library\\n");
    exit(2);
  }
  '''
    oldLibs = self.setCompilers.LIBS
    if self.haveLib('dl'):
      self.setCompilers.LIBS += ' -ldl'
    isShared = 0
    try:
      isShared = self.checkRun(defaultIncludes, body, defaultArg = defaultArg, executor = executor, timeout = timeout)
    except RuntimeError as e:
      if executor and str(e).find('Runaway process exceeded time limit') > -1:
        raise RuntimeError('Timeout: Unable to run MPI program with '+executor+'\n\
    (1) make sure this is the correct program to run MPI jobs\n\
    (2) your network may be misconfigured; see https://petsc.org/release/faq/#mpi-network-misconfigure\n\
    (3) you may have VPN running whose network settings may not play nice with MPI\n')

    self.setCompilers.LIBS = oldLibs
    if os.path.isfile(lib1Name) and self.framework.doCleanup: os.remove(lib1Name)
    if os.path.isfile(lib2Name) and self.framework.doCleanup: os.remove(lib2Name)
    if isShared:
      self.logPrint('Library was shared')
    else:
      self.logPrint('Library was not shared')
    return isShared

  def checkExportedSymbols(self, flags, checkLink = None, libraries = [], defaultArg = '', executor = None, timeout = 60):
    '''Determine whether an executable exports shared symbols
       - checkLink may be given as an alternative to the one in base.Configure'''
    exports = False

    if 'USE_VISIBILITY_C' in self.types.defines:
      visibility = '__attribute__((visibility ("default")))'
    else:
      visibility = ''

    # Make an executable that dynamically loads a symbol it contains
    guard = self.headers.getDefineName('dlfcn.h')
    if self.headers.headerPrefix:
      guard = self.headers.headerPrefix+'_'+guard
    defaultIncludes = '''
#include <stdio.h>
#include <stdlib.h>
#ifdef %s
#include <dlfcn.h>
#endif

#define PETSC_DLLEXPORT %s

extern PETSC_DLLEXPORT int foo(void) {
  return 42;
}
    ''' % (guard, visibility)
    body = '''
  void *lib;
  int (*foo)(void);

  lib = dlopen(NULL, RTLD_LAZY);
  if (!lib) {
    fprintf(stderr, "Could not open executable: %s\\n", dlerror());
    exit(1);
  }
  foo = (int (*)(void)) dlsym(lib, "foo");
  if (!foo) {
    fprintf(stderr, "Could not find function in executable\\n");
    exit(1);
  }
  if ((*foo)() != 42) {
    fprintf(stderr, "Could not run function\\n");
    exit(1);
  }
  '''
    oldFlags = self.setCompilers.CFLAGS
    oldLibs  = self.setCompilers.LIBS
    self.setCompilers.CFLAGS += ' '+flags
    if self.haveLib('dl'):
      self.setCompilers.LIBS += ' -ldl'
    try:
      exports = self.checkRun(defaultIncludes, body, defaultArg = defaultArg, executor = executor, timeout = timeout)
    except RuntimeError as e:
      self.logPrint('FAIL: '+str(e))
    self.setCompilers.CFLAGS = oldFlags
    self.setCompilers.LIBS   = oldLibs
    if exports:
      self.logPrint('Executable exports symbols for dlopen()')
    else:
      self.logPrint('Executable does not export symbols for dlopen()')
    return exports

  def isBGL(self):
    '''Returns true if compiler is IBM cross compiler for BGL'''
    if not hasattr(self, '_isBGL'):
      self.logPrint('**********Checking if running on BGL/IBM detected')
      if (self.check('', 'bgl_perfctr_void') or self.check('','ADIOI_BGL_Open')) and self.check('', '_xlqadd'):
        self.logPrint('*********BGL/IBM detected')
        self._isBGL = 1
      else:
        self.logPrint('*********BGL/IBM test failure')
        self._isBGL = 0
    return self._isBGL

  def checkPthreadMutex(self):
    '''Check for pthread mutex support'''
    funcs = ['pthread_mutex_unlock']
    prototypes = ['#include <pthread.h>']
    self.pthreadmutex = []
    calls = ['pthread_mutex_t m; pthread_mutex_unlock(&m)']
    if self.check('', funcs, prototype=prototypes, call=calls):
      self.logPrint('pthread mutex are linked in by default')
      self.pthreadmutex = []
      self.addDefine('HAVE_PTHREAD_MUTEX',1)
    elif self.check('pthread', funcs, prototype=prototypes, call=calls):
      self.logPrint('Using libpthread for the mutex')
      self.pthreadmutex = ['libpthread.a']
      self.addDefine('HAVE_PTHREAD_MUTEX',1)
    else:
      self.logPrint('No pthread mutex support found')
    return

  def checkExecutableExportFlag(self):
    '''Checks for the flag that allows executables to export symbols to dlsym()'''
    # Right now, we just check some compilers, but we should make a test trying to load a symbol from the executable
    # Discussion: https://stackoverflow.com/questions/6292473/how-to-call-function-in-executable-from-my-library/6298434#6298434
    for flag in ['', '-Wl,-export_dynamic', '-Wl,-export-dynamic', '-export-dynamic']:
      if self.checkExportedSymbols(flag):
        self.addDefine('HAVE_EXECUTABLE_EXPORT', 1)
        self.addMakeMacro('EXEFLAGS', flag)
        break
    return

  def configure(self):
    list(map(lambda args: self.executeTest(self.check, list(args)), self.libraries))
    self.executeTest(self.checkMath)
    self.executeTest(self.checkMathErf)
    self.executeTest(self.checkMathTgamma)
    self.executeTest(self.checkMathLgamma)
    self.executeTest(self.checkMathFenv)
    self.executeTest(self.checkMathLog2)
    self.executeTest(self.checkRealtime)
    self.executeTest(self.checkDynamic)
    self.executeTest(self.checkPthreadMutex)
    if not self.argDB['with-batch']:
      self.executeTest(self.checkExecutableExportFlag)
    return



#!/usr/bin/env python3

import os
petsc_hash_pkgs=os.path.join(os.getenv('HOME'),'petsc-hash-pkgs')

if __name__ == '__main__':
  import sys
  import os
  sys.path.insert(0, os.path.abspath('config'))
  import configure
  configure_options = [
    '--package-prefix-hash='+petsc_hash_pkgs,
    'CC=icx',
    'CXX=icpx',
    'FC=ifx',
    # Intel compilers enable GCC/clangs equivalent of -ffast-math *by default*. This is
    # bananas, so we make sure they use the same model as everyone else
    'COPTFLAGS=-g -O -fp-model=precise',
    'FOPTFLAGS=-g -O -fp-model=precise',
    'CXXOPTFLAGS=-g -O -fp-model=precise',
    '--with-blaslapack-dir='+os.environ['MKLROOT'],
    '--with-mkl_pardiso-dir='+os.environ['MKLROOT'],
    '--download-mpich=1',
    '--download-triangle=1',
    '--download-triangle-build-exec=1',
    '--download-ctetgen=1',
    '--download-tetgen=1',
    '--download-tetgen-build-exec=1',
    '--download-p4est=1',
    '--download-zlib=1',
    '--download-codipack=1',
    '--download-adblaslapack=1',
    '--download-kokkos',
    '--download-kokkos-cmake-arguments=-DKokkos_ENABLE_DEPRECATION_WARNINGS=OFF', # avoid warnings caused by broken [[deprecated]] in Intel compiler
    '--download-cmake', # need cmake-3.16+ to build Kokkos
    '--download-raja',
    '--with-strict-petscerrorcode',
  ]
  configure.petsc_configure(configure_options)



import config.package
import os

class Configure(config.package.Package):
  def __init__(self, framework):
    config.package.Package.__init__(self, framework)
    self.gitcommit         = 'dc552a8'
    self.download          = ['git://https://github.com/viennacl/viennacl-dev','https://github.com/viennacl/viennacl-dev/archive/'+self.gitcommit+'.tar.gz']
    self.downloaddirname   = [str('viennacl-dev')]
    self.includes          = ['viennacl/forwards.h']
    self.buildLanguages    = ['Cxx']
    self.downloadonWindows = 1
    self.complex           = 0
    self.devicePackage     = 1
    return

  def setupDependencies(self, framework):
    config.package.Package.setupDependencies(self, framework)
    self.cuda = framework.require('config.packages.cuda',self)
    self.opencl = framework.require('config.packages.opencl',self)
    self.openmp = framework.require('config.packages.openmp',self)
    self.deps = []
    self.odeps = [self.cuda,self.opencl,self.openmp]
    return

  def Install(self):
    import shutil
    import os
    self.log.write('ViennaCLDir = '+self.packageDir+' installDir '+self.installDir+'\n')
    #includeDir = self.packageDir
    srcdir     = os.path.join(self.packageDir, 'viennacl')
    destdir    = os.path.join(self.installDir, 'include', 'viennacl')
    try:
      if os.path.isdir(destdir): shutil.rmtree(destdir)
      shutil.copytree(srcdir,destdir)
    except RuntimeError as e:
      raise RuntimeError('Error installing ViennaCL include files: '+str(e))
    return self.installDir

  def configureLibrary(self):
    config.package.Package.configureLibrary(self)
    #check for CUDA:
    if not self.cuda.found:
      self.addDefine('HAVE_VIENNACL_NO_CUDA', 1)



import config.package

class Configure(config.package.Package):
  def __init__(self, framework):
    config.package.Package.__init__(self, framework)
    self.gitcommit = 'cbd530a38221043db9a0a4a127aca9cf477e50a5' #main july-04-2022
    self.download  = ['git://https://github.com/caidao22/cams.git','https://github.com/caidao22/cams/archive/'+self.gitcommit+'.tar.gz']
    self.functions = ['offline_cams_create']
    self.includes  = ['offline_schedule.h']
    self.liblist   = [['libcams.a']]
    self.hastests  = 1
    return

  def setupDependencies(self, framework):
    config.package.Package.setupDependencies(self, framework)
    return

  def Install(self):
    import os

    g = open(os.path.join(self.packageDir,'make.inc'),'w')
    g.write('CP               = '+self.programs.cp+'\n')
    g.write('RM               = '+self.programs.RM+'\n')
    g.write('MKDIR            = '+self.programs.mkdir+'\n')

    g.write('AR               = '+self.setCompilers.AR+'\n')
    g.write('ARFLAGS          = '+self.setCompilers.AR_FLAGS+'\n')
    g.write('AR_LIB_SUFFIX    = '+self.setCompilers.AR_LIB_SUFFIX+'\n')
    g.write('RANLIB           = '+self.setCompilers.RANLIB+'\n')

    g.write('PREFIX           = '+self.installDir+'\n')

    self.pushLanguage('Cxx')
    g.write('CXX               = '+self.getCompiler()+'\n')
    g.write('CXXFLAGS          = '+self.updatePackageCxxFlags(self.getCompilerFlags())+'\n')
    self.popLanguage()
    self.pushLanguage('C')
    g.write('CC                = '+self.getCompiler()+'\n')
    g.write('CFLAGS           = '+self.updatePackageCFlags(self.getCompilerFlags())+'\n')
    self.popLanguage()
    g.close()

    if self.installNeeded('make.inc'):
      self.logPrintBox('Configuring, compiling and installing cams; this may take several seconds')
      try:
        output1,err1,ret1 = config.package.Package.executeShellCommand(self.make.make_jnp_list + ['clean', 'lib'], cwd=self.packageDir, timeout=250, log=self.log)
      except RuntimeError as e:
        raise RuntimeError('Error running make on CAMS: '+str(e))
      try:
        output2,err2,ret2 = config.package.Package.executeShellCommand(self.make.make_jnp_list + ['install'], cwd=self.packageDir, timeout=250, log=self.log)
      except RuntimeError as e:
        raise RuntimeError('Error running install on CAMS: '+str(e))
      self.postInstall(output1+err1+output2+err2, 'make.inc')
    return self.installDir



import config.package

class Configure(config.package.Package):
  def __init__(self, framework):
    config.package.Package.__init__(self, framework)
    self.version           = '1.5.6'
    self.download          = ['https://github.com/facebook/zstd/archive/v'+self.version+'.tar.gz',
                              'https://web.cels.anl.gov/projects/petsc/download/externalpackages/zstd-'+self.version+'.tar.gz']
    self.functions         = ['ZSTD_compress']
    self.includes          = ['zstd.h']
    self.liblist           = [['libzstd.a']]
    self.downloaddirnames  = ['zstd']
    return

  def setupHelp(self, help):
    config.package.Package.setupHelp(self,help)
    import nargs
    return

  def setupDependencies(self, framework):
    config.package.Package.setupDependencies(self, framework)
    self.setCompilers = framework.require('config.setCompilers',self)
    self.make         = framework.require('config.packages.make',self)
    return

  def Install(self):
    import os
    with self.Language('C'):
      cc = self.getCompiler()
      cflags = self.getCompilerFlags()
    try:
      self.logPrintBox('Installing zstd; this may take several minutes')
      output,err,ret  = config.package.Package.executeShellCommand(self.make.make_jnp_list + ['CC='+cc, 'CFLAGS='+cflags, 'PREFIX='+self.installDir, 'install'], cwd=self.packageDir, timeout=250, log=self.log)
    except RuntimeError as e:
      raise RuntimeError('Error running make on zstd: '+str(e))
    return self.installDir



from __future__ import generators
import config.base

class Configure(config.base.Configure):
  def __init__(self, framework):
    config.base.Configure.__init__(self, framework)
    self.headerPrefix   = ''
    self.substPrefix    = ''
    self.have__fp16     = 0
    self.have__float128 = 0
    return

  def __str1__(self):
    output  = '  Scalar type: ' + self.scalartype + '\n'
    output += '  Precision: ' + self.precision + '\n'
    support = []
    if self.have__fp16 and not self.precision == '__fp16': support.append('__fp16')
    if self.have__float128 and not self.precision == '__float128': support.append('__float128')
    if not len(support) == 0: output += '  Support for ' + ' and '.join(support) + '\n'
    return output

  def setupHelp(self, help):
    import nargs
    #  Dec 2016, the __fp16 type is only available with GNU compilers on ARM systems
    help.addArgument('PETSc', '-with-precision=<__fp16,single,double,__float128>', nargs.Arg(None, 'double', 'Specify numerical precision'))
    help.addArgument('PETSc', '-with-scalar-type=<real or complex>', nargs.Arg(None, 'real', 'Specify real or complex numbers'))
    return

  def setupDependencies(self, framework):
    config.base.Configure.setupDependencies(self, framework)
    self.types         = framework.require('config.types', self)
    self.languages     = framework.require('PETSc.options.languages', self)
    self.compilers     = framework.require('config.compilers', self)
    self.libraries     = framework.require('config.libraries',self)
    self.setCompilers  = framework.require('config.setCompilers', self)
    self.compilerFlags = framework.require('config.compilerFlags', self)
    self.types         = framework.require('config.types', self)
    self.headers       = framework.require('config.headers', self)
    self.libraries     = framework.require('config.libraries', self)
    return

  def configureScalarType(self):
    '''Choose between real and complex numbers'''
    self.scalartype = self.framework.argDB['with-scalar-type'].lower()
    if self.scalartype == 'complex':
      self.addDefine('USE_COMPLEX', '1')
      if self.languages.clanguage == 'C' and not self.types.c99_complex:
        raise RuntimeError('C Compiler provided does not support C99 complex')
      if self.languages.clanguage == 'Cxx' and not self.types.cxx_complex:
        raise RuntimeError('Cxx compiler provided does not support std::complex')
      if self.languages.clanguage == 'Cxx':
        self.addDefine('USE_CXXCOMPLEX',1)
    elif not self.scalartype == 'real':
      raise RuntimeError('--with-scalar-type must be real or complex')
    self.logPrint('Scalar type is '+str(self.scalartype))
    # On apple isinf() and isnan() do not work when <complex> is included
    self.pushLanguage(self.languages.clanguage)
    if self.scalartype == 'complex' and self.languages.clanguage == 'Cxx':
      if self.checkLink('#include <math.h>\n#include <complex>\n','double b = 2.0;int a = isnormal(b);(void)a'):
        self.addDefine('HAVE_ISNORMAL',1)
      if self.checkLink('#include <math.h>\n#include <complex>\n','double b = 2.0;int a = isnan(b);(void)a'):
        self.addDefine('HAVE_ISNAN',1)
      elif self.checkLink('#include <float.h>\n#include <complex>\n','double b = 2.0;int a = _isnan(b);(void)a'):
        self.addDefine('HAVE__ISNAN',1)
      if self.checkLink('#include <math.h>\n#include <complex>\n','double b = 2.0;int a = isinf(b);(void)a'):
        self.addDefine('HAVE_ISINF',1)
      elif self.checkLink('#include <float.h>\n#include <complex>\n','double b = 2.0;int a = _finite(b);(void)a'):
        self.addDefine('HAVE__FINITE',1)
    else:
      if self.checkLink('#include <math.h>\n','double b = 2.0; int a = isnormal(b);(void)a'):
        self.addDefine('HAVE_ISNORMAL',1)
      if self.checkLink('#include <math.h>\n','double b = 2.0; int a = isnan(b);(void)a'):
        self.addDefine('HAVE_ISNAN',1)
      elif self.checkLink('#include <float.h>\n','double b = 2.0;int a = _isnan(b);(void)a'):
        self.addDefine('HAVE__ISNAN',1)
      if self.checkLink('#include <math.h>\n','double b = 2.0; int a = isinf(b);(void)a'):
        self.addDefine('HAVE_ISINF',1)
      elif self.checkLink('#include <float.h>\n','double b = 2.0;int a = _finite(b);(void)a'):
        self.addDefine('HAVE__FINITE',1)
    self.popLanguage()
    return

  def checkNoFiniteMathOnly(self):
    '''Check if attribute for ignoring finite-math-only optimization is valid, for isnan() and isinf()'''
    if self.checkCompile('','__attribute__((optimize ("no-finite-math-only"))) int foo(void);'):
      self.addDefine('HAVE_NO_FINITE_MATH_ONLY',1)

  def configurePrecision(self):
    '''Set the default real number precision for PETSc objects'''
    self.log.write('Checking C compiler works with __float128\n')
    self.have__float128 = 0
    if self.libraries.check('quadmath','logq',prototype='#include <quadmath.h>',call='__float128 f = 0.0; logq(f)'):
      self.log.write('C compiler works with quadmath library\n')
      self.have__float128 = 1
      if hasattr(self.compilers, 'FC'):
        self.libraries.pushLanguage('FC')
        self.log.write('Checking Fortran compiler works with quadmath library\n')
        if self.libraries.check('quadmath','     ',call = '      real*16 s,w; w = 2.0; s = cos(w)'):
          self.log.write('Fortran compiler works with quadmath library\n')
        else:
          self.have__float128 = 0
          self.log.write('Fortran compiler fails with quadmath library\n')
        self.libraries.popLanguage()
      if hasattr(self.compilers, 'CXX'):
        self.libraries.pushLanguage('Cxx')
        isGNU = self.setCompilers.isGNU(self.getCompiler(lang='Cxx'), self.log)
        # need to bypass g++ error: non-standard suffix on floating constant [-Werror=pedantic]
        # this warning can't be disabled but is actually never triggered by PETSc
        if isGNU:
          self.setCompilers.pushLanguage('Cxx')
          preprocessorFlagsArg = self.setCompilers.getPreprocessorFlagsArg()
          oldPreprocessorFlags = getattr(self.setCompilers, preprocessorFlagsArg)
          setattr(self.setCompilers, preprocessorFlagsArg, oldPreprocessorFlags+' -Wno-error')
        self.log.write('Checking C++ compiler works with quadmath library\n')
        if self.libraries.check('quadmath','logq',prototype='#include <quadmath.h>',call='__float128 f = FLT128_EPSILON; logq(f)'):
          self.log.write('C++ compiler works with quadmath library\n')
        else:
          self.have__float128 = 0
          self.log.write('C++ compiler fails with quadmath library\n')
        if isGNU:
          setattr(self.setCompilers, preprocessorFlagsArg, oldPreprocessorFlags)
          self.setCompilers.popLanguage()
        self.libraries.popLanguage()
      if self.have__float128:
          self.libraries.add('quadmath','logq',prototype='#include <quadmath.h>',call='__float128 f = 0.0; logq(f)')
          self.addDefine('HAVE_REAL___FLOAT128', '1')

    self.log.write('Checking C compiler works with __fp16\n')
    self.have__fp16 = 0
    if self.libraries.check('','',call='__fp16 f = 1.0, g; g = ret___fp16(f); (void)g',prototype='static __fp16 ret___fp16(__fp16 f) { return f; }'):
      self.have__fp16 = 1
      self.addDefine('HAVE_REAL___FP16', '1')

    self.precision = self.framework.argDB['with-precision'].lower()
    if self.precision == '__fp16':  # supported by gcc trunk
      if self.scalartype == 'complex':
        raise RuntimeError('__fp16 can only be used with real numbers, not complex')
      if hasattr(self.compilers, 'FC'):
        raise RuntimeError('__fp16 can only be used with C compiler, not Fortran')
      if self.have__fp16:
        self.addDefine('USE_REAL___FP16', '1')
        self.addMakeMacro('PETSC_SCALAR_SIZE', '16')
      else:
        raise RuntimeError('__fp16 support not found, cannot proceed --with-precision=__fp16')
    elif self.precision == 'single':
      self.addDefine('USE_REAL_SINGLE', '1')
      self.addMakeMacro('PETSC_SCALAR_SIZE', '32')
    elif self.precision == 'double':
      self.addDefine('USE_REAL_DOUBLE', '1')
      self.addMakeMacro('PETSC_SCALAR_SIZE', '64')
    elif self.precision == '__float128':  # supported by gcc 4.6/gfortran and later
      if self.have__float128:
        self.addDefine('USE_REAL___FLOAT128', '1')
        self.addMakeMacro('PETSC_SCALAR_SIZE', '128')
      else:
        raise RuntimeError('__float128 support not found. --with-precision=__float128 works with gcc-4.6 and newer compilers.')
    else:
      raise RuntimeError('--with-precision must be __fp16, single, double, or __float128')
    self.logPrint('Precision is '+str(self.precision))
    return

  def configure(self):
    self.executeTest(self.configureScalarType)
    self.executeTest(self.configurePrecision)
    self.executeTest(self.checkNoFiniteMathOnly)
    return

  def precisionToBytes(self):
    d = {'__fp16' : 2, 'single' : 4, 'double' : 8, '__float128' : 16}
    return d[self.precision]



import config.package
import os

class Configure(config.package.CMakePackage):
  def __init__(self, framework):
    config.package.CMakePackage.__init__(self, framework)
    self.version          = '7.0.0'
    self.minversion       = '5.2.1' # bugs in 5.2.0 prevent it from functioning
    self.versionname      = 'SUPERLU_MAJOR_VERSION.SUPERLU_MINOR_VERSION.SUPERLU_PATCH_VERSION'
    #self.gitcommit        = 'v'+self.version
    self.gitcommit        = 'b814567e6cdeba61edffe45839a4043a46215c7a' # master feb-26-2024
    self.download         = ['git://https://github.com/xiaoyeli/superlu','https://github.com/xiaoyeli/superlu/archive/'+self.gitcommit+'.tar.gz']
    self.functions        = ['set_default_options']
    self.includes         = ['slu_ddefs.h']
    self.liblist          = [['libsuperlu.a']]
    # SuperLU has NO support for 64-bit integers, use SuperLU_DIST if you need that
    self.requires32bitint = 1;  # 1 means that the package will not work with 64-bit integers
    self.excludedDirs     = ['superlu_dist','superlu_mt']
    # SuperLU does not work with --download-fblaslapack with Compaqf90 compiler on windows.
    # However it should work with intel ifort.
    self.downloadonWindows= 1
    self.hastests         = 1
    self.hastestsdatafiles= 1
    self.precisions       = ['single','double']
    return

  def setupDependencies(self, framework):
    config.package.CMakePackage.setupDependencies(self, framework)
    self.blasLapack = self.framework.require('config.packages.BlasLapack',self)
    self.deps       = [self.blasLapack]
    return

  def formCMakeConfigureArgs(self):
    args = config.package.CMakePackage.formCMakeConfigureArgs(self)
    args.append('-DUSE_XSDK_DEFAULTS=YES')
    args.append('-DCMAKE_DISABLE_FIND_PACKAGE_Doxygen=TRUE')
    args.append('-DTPL_BLAS_LIBRARIES="'+self.libraries.toString(self.blasLapack.dlib)+'"')

    #  Tests are broken on Apple since they depend on a shared library that is not resolved against BLAS
    args.append('-Denable_tests=0')
    args.append('-Denable_examples=0')

    if not hasattr(self.compilers, 'FC'):
      args.append('-DXSDK_ENABLE_Fortran=OFF')

    # Add in fortran mangling flag
    if self.blasLapack.mangling == 'underscore':
      mangledef = '-DAdd_'
    elif self.blasLapack.mangling == 'caps':
      mangledef = '-DUpCase'
    else:
      mangledef = '-DNoChange'
    for place,item in enumerate(args):
      if item.find('CMAKE_C_FLAGS') >= 0 or item.find('CMAKE_CXX_FLAGS') >= 0:
        args[place]=item[:-1]+' '+mangledef+'"'

    return args




import config.package
import os

class Configure(config.package.GNUPackage):
  def __init__(self, framework):
    config.package.GNUPackage.__init__(self, framework)
    self.gitcommit         = 'openssl-3.0.8'
    self.download          = ['git://https://github.com/openssl/openssl.git','https://github.com/openssl/openssl/archive/'+self.gitcommit+'.tar.gz']
    self.functions         = ['SSL_version']
    self.includes          = ['openssl/ssl.h']
    self.liblist           = [['libssl.a','libcrypto.a']]
    self.configureName     = 'Configure'

  def setupHelp(self, help):
    import nargs
    config.package.GNUPackage.setupHelp(self, help)
    help.addArgument('SSL', '-with-ssl-certificate=<bool>',nargs.ArgBool(None, 0, 'Require certificate with SSL'))

  def setupDependencies(self, framework):
    config.package.GNUPackage.setupDependencies(self, framework)
    self.deps = []

  def getSearchDirectories(self):
    '''macOS no longer provides openssl include files. On macOS brew puts them in the second location listed here'''
    return ['',os.path.join('/usr','local','opt','openssl@1.1'),os.path.join('/usr','local','opt','openssl')]

  def formGNUConfigureArgs(self):
    '''This sets up the prefix, compiler flags, shared flags, and other generic arguments
       that are fed into the configure script supplied with the package.
       Override this to set options needed by a particular package'''
    args=[]
    ## prefix
    args.append('--prefix='+self.installDir)
    args.append('MAKE='+self.make.make)
    args.append('--libdir='+self.libDir)
    ## compiler args
    self.pushLanguage('C')
    if not self.installwithbatch and hasattr(self.setCompilers,'cross_cc'):
      args.append('CC="'+self.setCompilers.cross_cc+'"')
    else:
      args.append('CC="'+self.getCompiler()+'"')
    args.append('CFLAGS="'+self.updatePackageCFlags(self.getCompilerFlags())+'"')
    args.append('AR="'+self.setCompilers.AR+'"')
    args.append('ARFLAGS="'+self.setCompilers.AR_FLAGS+'"')
    if not self.installwithbatch and hasattr(self.setCompilers,'cross_LIBS'):
      args.append('LIBS="'+self.setCompilers.cross_LIBS+'"')
    if self.setCompilers.LDFLAGS:
      args.append('LDFLAGS="'+self.setCompilers.LDFLAGS+'"')
    self.popLanguage()
    return args

  def configureLibrary(self):
    if 'with-ios' in self.argDB and self.argDB['with-ios']:
      self.found = 0
      return
    config.package.GNUPackage.configureLibrary(self)

  def consistencyChecks(self):
   config.package.GNUPackage.consistencyChecks(self)
   if self.argDB['with-'+self.package]:
     if self.argDB['with-ssl-certificate']:
       self.addDefine('USE_SSL_CERTIFICATE','1')



#!/usr/bin/env python3
if __name__ == '__main__':
  import sys
  import os
  sys.path.insert(0, os.path.abspath('config'))
  import configure
  configure_options = [
    '--with-make-test-np=20',
    'COPTFLAGS=-g -O',
    'FOPTFLAGS=-g -O',
    'CXXOPTFLAGS=-g -O',
    '--download-viennacl',
    '--with-opencl-include=/usr/local/cuda/include',
    '--with-opencl-lib=-L/usr/local/cuda/lib64 -lOpenCL',
    '--with-cuda=0',
    '--download-codipack=1',
    '--download-adblaslapack=1',
    '--with-strict-petscerrorcode',
    '--with-coverage',
  ]
  configure.petsc_configure(configure_options)



import config.package
import os

class Configure(config.package.GNUPackage):
  def __init__(self, framework):
    config.package.GNUPackage.__init__(self, framework)
    self.versionname       = "GSL_MAJOR_VERSION.GSL_MINOR_VERSION"
    self.download          = ['ftp://ftp.gnu.org/gnu/gsl/gsl-2.7.1.tar.gz']
    self.functions         = ['gsl_sf_hermite_zero']
    self.includes          = ['gsl/gsl_version.h']
    self.liblist           = [['libgsl.a','libgslcblas.a']]
    self.downloadonWindows = 1
    return

  def setupDependencies(self, framework):
    config.package.Package.setupDependencies(self, framework)
    self.mathlib         = framework.require('config.packages.mathlib',self)
    self.deps            = [self.mathlib]
    return

  def Install(self):
    macos_deployment = ''
    if 'MACOSX_DEPLOYMENT_TARGET' in os.environ:
      macos_deployment = os.environ['MACOSX_DEPLOYMENT_TARGET']
      msg = 'WARNING! Found environment variable: %s=%s\n' % ('MACOSX_DEPLOYMENT_TARGET', os.environ['MACOSX_DEPLOYMENT_TARGET'])
      self.logPrintBox(msg+'Removing it for GSL build, since it breaks the GSL build')
      del os.environ['MACOSX_DEPLOYMENT_TARGET']
    installDir = config.package.GNUPackage.Install(self)
    if macos_deployment:
      os.environ['MACOSX_DEPLOYMENT_TARGET'] = macos_deployment
    return installDir






#!/usr/bin/env python3

import os

configure_options = [
  '--with-cc=emcc',
  '--with-cxx=0',
  '--with-fc=0',
  '--with-ranlib=emranlib',
  '--with-ar=emar',
  '--with-shared-libraries=0',
  '--download-f2cblaslapack=1',
  '--with-mpi=0',
  '--with-batch',
  '--with-strict-petscerrorcode',
  ]

if __name__ == '__main__':
  import sys,os
  sys.path.insert(0,os.path.abspath('config'))
  import configure
  configure.petsc_configure(configure_options)



import config.package
import os
import script

class Configure(config.package.Package):
  def __init__(self, framework):
    config.package.Package.__init__(self, framework)
    self.download          = ['https://github.com/mpi4py/mpi4py/releases/download/4.0.3/mpi4py-4.0.3.tar.gz']
    self.functions         = []
    self.includes          = []
    self.useddirectly      = 0
    return

  def setupDependencies(self, framework):
    config.package.Package.setupDependencies(self, framework)
    self.python          = framework.require('config.packages.python',self)
    self.setCompilers    = framework.require('config.setCompilers',self)
    self.sharedLibraries = framework.require('PETSc.options.sharedLibraries', self)
    self.installdir      = framework.require('PETSc.options.installDir',self)
    self.mpi             = framework.require('config.packages.MPI',self)
    self.deps            = [self.mpi]
    return

  def Install(self):
    installLibPath = os.path.join(self.installDir, 'lib')
    if self.setCompilers.isDarwin(self.log):
      apple = 'You may need to\n (csh/tcsh) setenv MACOSX_DEPLOYMENT_TARGET 10.X\n (sh/bash) MACOSX_DEPLOYMENT_TARGET=10.X; export MACOSX_DEPLOYMENT_TARGET\nbefore running make on PETSc'
    else:
      apple = ''
    self.logClearRemoveDirectory()
    self.logResetRemoveDirectory()
    archflags = ""
    if self.setCompilers.isDarwin(self.log):
      if self.setCompilers.isARM(self.log):
        archflags = "ARCHFLAGS=\'-arch arm64\' "
      elif self.types.sizes['void-p'] == 4:
        archflags = "ARCHFLAGS=\'-arch i386\' "
      else:
        archflags = "ARCHFLAGS=\'-arch x86_64\' "

    self.framework.pushLanguage('C')
    self.logPrintBox('Building mpi4py, this may take several minutes')
    cleancmd = 'MPICC='+self.framework.getCompiler()+'  '+archflags+self.python.pyexe+' setup.py clean --all  2>&1'
    output,err,ret  = config.base.Configure.executeShellCommand(cleancmd, cwd=self.packageDir,  checkCommand=script.Script.passCheckCommand, timeout=100, log=self.log)
    if ret: raise RuntimeError('Error cleaning mpi4py. Check configure.log')

    cflags = ''
    # by default, multiple flags are added by setup.py (-DNDEBUG -O3 -g), no matter the type of PETSc build
    # this is problematic with Intel compilers, which take extremely long to compile bindings when using -g
    # so we instead force no additional flags (other than the ones already used by PETSc, i.e., CFLAGS)
    # TODO FIXME: this observation was made with Intel(R) oneAPI DPC++/C++ Compiler 2025.1.0 (2025.1.0.20250317), but it may be fixed in a subsequent release
    if config.setCompilers.Configure.isIntel(self.getCompiler(), self.log):
      cflags = 'CFLAGS=\''+self.getCompilerFlags()+'\' '
    buildcmd = 'MPICC='+self.framework.getCompiler()+'  '+archflags+cflags+self.python.pyexe+' setup.py build 2>&1'
    output,err,ret  = config.base.Configure.executeShellCommand(buildcmd, cwd=self.packageDir, checkCommand=script.Script.passCheckCommand,timeout=100, log=self.log)
    if ret: raise RuntimeError('Error building mpi4py. Check configure.log')

    self.logPrintBox('Installing mpi4py')
    installcmd = 'MPICC='+self.framework.getCompiler()+' '+self.python.pyexe+' setup.py install --install-lib='+installLibPath+' 2>&1'
    output,err,ret  = config.base.Configure.executeShellCommand(installcmd, cwd=self.packageDir,checkCommand=script.Script.passCheckCommand, timeout=100, log=self.log)
    if ret: raise RuntimeError('Error installing mpi4py. Check configure.log')
    self.framework.popLanguage()
    return self.installDir

  def configureLibrary(self):
    self.checkDownload()
    if not self.sharedLibraries.useShared:
        raise RuntimeError('mpi4py requires PETSc be built with shared libraries; rerun with --with-shared-libraries')
    if not getattr(self.python,'numpy'):
        raise RuntimeError('mpi4py, in the context of PETSc,requires Python with numpy module installed.\n'
                           'Please install using package managers - for ex: "apt" or "dnf" (on linux),\n'
                           'or  using: %s -m pip install %s' % (self.python.pyexe, 'numpy'))
    if self.argDB.get('with-mpi4py-dir'):
      self.directory = self.argDB['with-mpi4py-dir']
    elif self.argDB.get('download-mpi4py'):
      self.directory = os.path.join(self.installDir)
    elif self.argDB.get('with-mpi4py'):
      if not getattr(self.python,'mpi4py'):
        raise RuntimeError('mpi4py not found in default Python PATH! Suggest using --download-mpi4py!')
    else:
        raise RuntimeError('mpi4py unreconginzed mode of building mpi4py! Suggest using --download-mpi4py!')

    if self.directory:
      installLibPath = os.path.join(self.directory, 'lib')
      if not os.path.isfile(os.path.join(installLibPath,'mpi4py','__init__.py')):
        raise RuntimeError('mpi4py not found at %s' % installLibPath)
      self.addMakeMacro('PETSC_MPI4PY_PYTHONPATH',installLibPath)
      if 'PYTHONPATH' in os.environ:
        self.logPrintBox('To use mpi4py, do\nexport PYTHONPATH=${PYTHONPATH}'+os.pathsep+installLibPath,rmDir = 0)
        os.environ['PYTHONPATH'] = os.environ['PYTHONPATH']+os.pathsep+installLibPath
      else:
        self.logPrintBox('To use mpi4py, do\nexport PYTHONPATH='+installLibPath,rmDir = 0)
        os.environ['PYTHONPATH'] = installLibPath

    self.addMakeMacro('MPI4PY',"yes")
    self.found = 1




#!/usr/bin/env python3

"""
PETSc: Portable, Extensible Toolkit for Scientific Computation
==============================================================

The Portable, Extensible Toolkit for Scientific Computation (PETSc),
is a suite of data structures and routines for the scalable (parallel)
solution of scientific applications modeled by partial differential
equations. It employs the Message Passing Interface (MPI) standard for
all message-passing communication.

.. note::

   To install the ``PETSc`` and ``petsc4py`` packages use::

     $ python -m pip install numpy
     $ python -m pip install petsc petsc4py

.. tip::

  You can also install the in-development versions with::

    $ python -m pip install cython numpy
    $ python -m pip install --no-deps https://gitlab.com/petsc/petsc/-/archive/main/petsc-main.tar.gz

  Provide any ``PETSc`` ``./configure`` options using the environmental variable ``PETSC_CONFIGURE_OPTIONS``.

  Do not use the ``PETSc`` ``./configure`` options ``--with-cc``, ``--with-cxx``, ``--with-fc``, or ``--with-mpi-dir``.
  Compilers are detected from (in order): 1) the environmental variables ``MPICC``, ``MPICXX``, ``MPIFORT``,
  or 2) from running the ``which mpicc``, ``which mpicxx``, ``which mpifort`` commands.

"""

import re
import os
import sys
import shlex
import shutil
from setuptools import setup
from setuptools.command.install import install as _install
try:
    from setuptools.command.bdist_wheel import bdist_wheel as _bdist_wheel
except ImportError:
    from wheel.bdist_wheel import bdist_wheel as _bdist_wheel
from distutils import log

init_py = """\
# Author:  PETSc Team
# Contact: petsc-maint@mcs.anl.gov

def get_petsc_dir():
    import os
    return os.path.dirname(__file__)


def get_config():
    conf = {}
    conf['PETSC_DIR'] = get_petsc_dir()
    return conf
"""

main_py = """\
# Author:  PETSc Team
# Contact: petsc-maint@mcs.anl.gov

if __name__ == "__main__":
    import sys
    if "--prefix" in sys.argv:
        from . import get_petsc_dir
        print(get_petsc_dir())
        del get_petsc_dir
    del sys
"""

metadata = {
    'provides' : ['petsc'],
    'zip_safe' : False,
}

CONFIGURE_OPTIONS = []


def bootstrap():
    # Set PETSC_DIR and PETSC_ARCH
    PETSC_DIR  = os.path.abspath(os.getcwd())
    PETSC_ARCH = 'arch-python'
    os.environ['PETSC_DIR']  = PETSC_DIR
    os.environ['PETSC_ARCH'] = PETSC_ARCH
    sys.path.insert(0, os.path.join(PETSC_DIR, 'config'))
    sys.path.insert(0, os.path.join(PETSC_DIR, 'lib','petsc','conf'))

    # Generate package __init__.py and __main__.py files
    pkgdir = os.path.join('config', 'pypi')
    os.makedirs(pkgdir, exist_ok=True)
    for pyfile, contents in (
        ('__init__.py', init_py),
        ('__main__.py', main_py),
    ):
        with open(os.path.join(pkgdir, pyfile), 'w') as fh:
            fh.write(contents)

    # Configure options
    options = os.environ.get('PETSC_CONFIGURE_OPTIONS', '')
    CONFIGURE_OPTIONS.extend(shlex.split(options))
    for i in CONFIGURE_OPTIONS:
        if i.startswith('--with-mpi-dir='):
            raise RuntimeError("Do not use --with-mpi-dir, use the environmental variables MPICC, MPICXX, MPIFORT")
        if i.startswith('--with-cc='):
            raise RuntimeError("Do not use --with-cc, use the environmental variable MPICC")
        if i.startswith('--with-cxx=') and i != "--with-cxx=0":
            raise RuntimeError("Do not use --with-cxx, use the environmental variable MPICXX")
        if i.startswith('--with-fc=') and i != "--with-fc=0":
            raise RuntimeError("Do not use --with-fc, use the environmental variable MPIFORT")



def config(prefix, dry_run=False):
    log.info('PETSc: configure')
    options = [
        '--prefix=' + prefix,
        'PETSC_ARCH='+os.environ['PETSC_ARCH'],
        '--with-shared-libraries=1',
        '--with-c2html=0', # not needed
        ]
    if '--with-fc=0' in CONFIGURE_OPTIONS:
        options.append('--with-sowing=0')
    if '--with-debugging=1' not in CONFIGURE_OPTIONS:
        options.append('--with-debugging=0')
    if '--with-mpi=0' not in CONFIGURE_OPTIONS:
        mpicc = os.environ.get('MPICC') or shutil.which('mpicc')
        mpicxx = os.environ.get('MPICXX') or shutil.which('mpicxx')
        mpifort = (
            os.environ.get('MPIFORT')
            or os.environ.get('MPIF90')
            or shutil.which('mpifort')
            or shutil.which('mpif90')
        )
        if mpicc:
            options.append('--with-cc='+mpicc)
            if '--with-cxx=0' not in CONFIGURE_OPTIONS:
                if mpicxx:
                    options.append('--with-cxx='+mpicxx)
                else:
                    options.append('--with-cxx=0')
            if '--with-fc=0' not in CONFIGURE_OPTIONS:
                if mpifort:
                    options.append('--with-fc='+mpifort)
                else:
                    options.append('--with-fc=0')
                    options.append('--with-sowing=0')
        else:
            options.append('--with-mpi=0')
    options.extend(CONFIGURE_OPTIONS)
    #
    log.info('configure options:')
    for opt in options:
        log.info(' '*4 + opt)
    # Run PETSc configure
    if dry_run:
        return
    use_config_py = False
    if use_config_py:
        import configure
        configure.petsc_configure(options)
        import logger
        logger.Logger.defaultLog = None
    else:
        python = sys.executable
        command = [python, './configure'] + options
        status = os.system(" ".join(command))
        if status != 0:
            raise RuntimeError(status)

    # Fix PETSc configuration
    using_build_backend = any(
        os.environ.get(prefix + '_BUILD_BACKEND')
        for prefix in ('_PYPROJECT_HOOKS', 'PEP517')
    )
    if using_build_backend:
        pdir = os.environ['PETSC_DIR']
        parch = os.environ['PETSC_ARCH']
        include = os.path.join(pdir, parch, 'include')
        for filename in (
            'petscconf.h',
            'petscconfiginfo.h',
            'petscmachineinfo.h',
        ):
            filename = os.path.join(include, filename)
            with open(filename, 'r') as old_fh:
                contents = old_fh.read()
            contents = contents.replace(prefix, '${PETSC_DIR}')
            contents = re.sub(
                r'^(#define PETSC_PYTHON_EXE) "(.*)"$',
                r'\1 "python%d"' % sys.version_info[0],
                contents, flags=re.MULTILINE,
            )
            with open(filename, 'w') as new_fh:
                new_fh.write(contents)


def build(dry_run=False):
    log.info('PETSc: build')
    # Run PETSc build
    if dry_run:
        return
    use_builder_py = False
    if use_builder_py:
        import builder
        builder.PETScMaker().run()
        import logger
        logger.Logger.defaultLog = None
    else:
        make = shutil.which('make')
        command = [make, 'all']
        status = os.system(" ".join(command))
        if status != 0:
            raise RuntimeError(status)


def install(dry_run=False):
    log.info('PETSc: install')
    # Run PETSc installer
    if dry_run:
        return
    use_install_py = False
    if use_install_py:
        import install
        install.Installer().run()
        import logger
        logger.Logger.defaultLog = None
    else:
        make = shutil.which('make')
        command = [make, 'install']
        status = os.system(" ".join(command))
        if status != 0:
            raise RuntimeError(status)


class context(object):
    def __init__(self):
        self.sys_argv = sys.argv[:]
        self.wdir = os.getcwd()
    def enter(self):
        del sys.argv[1:]
        pdir = os.environ['PETSC_DIR']
        os.chdir(pdir)
        return self
    def exit(self):
        sys.argv[:] = self.sys_argv
        os.chdir(self.wdir)


class cmd_install(_install):

    def initialize_options(self):
        _install.initialize_options(self)

    def finalize_options(self):
        _install.finalize_options(self)
        self.install_lib = self.install_platlib
        self.install_libbase = self.install_lib
        self.old_and_unmanageable = True

    def run(self):
        root_dir = os.path.abspath(self.install_lib)
        prefix = os.path.join(root_dir, 'petsc')
        #
        ctx = context().enter()
        try:
            config(prefix, self.dry_run)
            build(self.dry_run)
            install(self.dry_run)
        finally:
            ctx.exit()
        #
        self.outputs = []
        for dirpath, _, filenames in os.walk(prefix):
            for fn in filenames:
                self.outputs.append(os.path.join(dirpath, fn))
        #
        _install.run(self)

    def get_outputs(self):
        outputs = getattr(self, 'outputs', [])
        outputs += _install.get_outputs(self)
        return outputs


class cmd_bdist_wheel(_bdist_wheel):

    def finalize_options(self):
        super().finalize_options()
        self.root_is_pure = False
        self.build_number = None
        # self.keep_temp = True

    def get_tag(self):
        plat_tag = super().get_tag()[-1]
        return (self.python_tag, "none", plat_tag)


def version():
    version_re = {
        'major'  : re.compile(r"#define\s+PETSC_VERSION_MAJOR\s+(\d+)"),
        'minor'  : re.compile(r"#define\s+PETSC_VERSION_MINOR\s+(\d+)"),
        'micro'  : re.compile(r"#define\s+PETSC_VERSION_SUBMINOR\s+(\d+)"),
        'release': re.compile(r"#define\s+PETSC_VERSION_RELEASE\s+([-]*\d+)"),
        }
    petscversion_h = os.path.join('include','petscversion.h')
    data = open(petscversion_h, 'r').read()
    major = int(version_re['major'].search(data).groups()[0])
    minor = int(version_re['minor'].search(data).groups()[0])
    micro = int(version_re['micro'].search(data).groups()[0])
    release = int(version_re['release'].search(data).groups()[0])
    if release:
        v = "%d.%d.%d" % (major, minor, micro)
    else:
        v = "%d.%d.0.dev%d" % (major, minor+1, 0)
    return v


def tarball():
    VERSION = version()
    if '.dev' in VERSION:
        return None
    return ('https://web.cels.anl.gov/projects/petsc/download/release-snapshots/'
            'petsc-%s.tar.gz#egg=petsc-%s' % (VERSION, VERSION))


description = __doc__.split('\n')[1:-1]
del description[1:3]

classifiers = """
Development Status :: 5 - Production/Stable
Intended Audience :: Developers
Intended Audience :: Science/Research
Operating System :: POSIX
Programming Language :: C
Programming Language :: C++
Programming Language :: Fortran
Programming Language :: Python
Programming Language :: Python :: 3
Topic :: Scientific/Engineering
Topic :: Software Development :: Libraries
"""

bootstrap()
setup(
    name='petsc',
    version=version(),
    description=description.pop(0),
    long_description='\n'.join(description),
    long_description_content_type='text/x-rst',
    classifiers=classifiers.split('\n')[1:-1],
    keywords = ['PETSc', 'MPI'],
    platforms=['POSIX'],
    license='BSD-2-Clause',

    url='https://petsc.org/',
    download_url=tarball(),

    author='PETSc Team',
    author_email='petsc-maint@mcs.anl.gov',
    maintainer='Lisandro Dalcin',
    maintainer_email='dalcinl@gmail.com',

    packages=['petsc'],
    package_dir= {'petsc': 'config/pypi'},
    cmdclass={
        'install': cmd_install,
        'bdist_wheel': cmd_bdist_wheel,
    },
    **metadata
)



import config.package

class Configure(config.package.GNUPackage):
  def __init__(self, framework):
    config.package.GNUPackage.__init__(self, framework)
    self.version           = '2.1.1'
    self.versionname       = 'SZLIB_VERSION'
    self.download          = ['https://support.hdfgroup.org/ftp/lib-external/szip/'+self.version+'/src/szip-2.1.1.tar.gz',
                              'https://web.cels.anl.gov/projects/petsc/download/externalpackages/szip-'+self.version+'.tar.gz']
    self.functions         = ['SZ_BufftoBuffCompress', 'SZ_BufftoBuffDecompress']
    self.includes          = ['szlib.h']
    self.liblist           = [['libsz.a'],['szlib.lib']]
    self.downloaddirnames  = ['szip']
    return




#!/usr/bin/python3

#  Modules loaded by default (on login to spock):
#
#  1) craype-x86-rome                          8) cce/12.0.3
#  2) libfabric/1.11.0.4.75                    9) craype/2.7.11
#  3) craype-network-ofi                      10) cray-dsmml/0.2.2
#  4) perftools-base/21.10.0                  11) cray-mpich/8.1.10
#  5) xpmem/2.2.40-2.1_2.44__g3cf3325.shasta  12) cray-libsci/21.08.1.2
#  6) cray-pmi/6.0.14                         13) PrgEnv-cray/8.2.0
#  7) cray-pmi-lib/6.0.14                     14) DefApps/default
#
# Need to load additional rocm module to build with hip
#
# module load rocm/4.3.0
#
# Note: LIBS option below is needed to use GPU enabled MPI. It also requires
# the following env variables to be set at runtime
#
# export MPIR_CVAR_GPU_EAGER_DEVICE_MEM=0
# export MPICH_GPU_SUPPORT_ENABLED=1
# export MPICH_SMP_SINGLE_COPY_MODE=CMA
#
# Additional note: If "craype-accel-amd-gfx908" module is loaded (that is
# needed for "OpenMP offload") - it causes link errors when using 'cc or hipcc'
# with fortran objs, hence not used
#

if __name__ == '__main__':
  import sys
  import os
  sys.path.insert(0, os.path.abspath('config'))
  import configure
  configure_options = [
    '--with-cc=cc',
    '--with-cxx=CC',
    '--with-fc=ftn',
    'LIBS=-L{x}/gtl/lib -lmpi_gtl_hsa'.format(x=os.environ['CRAY_MPICH_ROOTDIR']),
    '--with-debugging=0',
    '--with-mpiexec=srun -p ecp -N 1 -A csc314 -t 00:10:00',
    '--with-hip=1',
    '--with-hipc=hipcc',
    '--download-kokkos=1',
    '--download-kokkos-kernels=1',
    '--download-magma=1',
    '--with-magma-gputarget=gfx908',
  ]
  configure.petsc_configure(configure_options)



import config.package
import os

class Configure(config.package.CMakePackage):
  def __init__(self, framework):
    config.package.CMakePackage.__init__(self, framework)
    self.gitcommit        = '65328ea'
    self.download         = ['git://https://github.com/MeshToolkit/MSTK','https://github.com/MeshToolkit/MSTK/archive'+self.gitcommit+'.tar.gz']
    self.downloaddirnames = ['mstk']
    self.includes         = ['MSTK.h']
    self.liblist          = [['libmstk.a']]
    self.functions        = []
    self.buildLanguages   = ['Cxx']
    self.hastests         = 1
    return

  def setupDependencies(self, framework):
    config.package.CMakePackage.setupDependencies(self, framework)
    self.compilerFlags   = framework.require('config.compilerFlags', self)
    self.mpi             = framework.require('config.packages.MPI',self)
    self.metis           = framework.require('config.packages.metis',self)
    self.parmetis        = framework.require('config.packages.parmetis',self)
    self.ptscotch        = framework.require('config.packages.PTScotch',self)
    self.zoltan          = framework.require('config.packages.Zoltan',self)
    self.exodusii        = framework.require('config.packages.exodusii',self)
    self.trilinos        = framework.require('config.packages.Trilinos',self)
    self.deps            = [self.mpi]
    return

  def formCMakeConfigureArgs(self):
    if self.checkSharedLibrariesEnabled():
      raise RuntimeError('mstk cannot be built with shared libraries, run with --download-mstk-shared=0')

    args = config.package.CMakePackage.formCMakeConfigureArgs(self)
    args.append('-DUSE_XSDK_DEFAULTS=YES')
    if self.compilerFlags.debugging:
      args.append('-DXSDK_ENABLE_DEBUG=YES')
    else:
      args.append('-DXSDK_ENABLE_DEBUG=NO')

    args.append('-DENABLE_PARALLEL=yes')
    if not self.metis.found and not self.zoltan.found and not self.trilinos:
      raise  RuntimeError('MSTK requires either Metis (--download-metis) or Zoltan (--download-zoltan (or --download-trilinos) --download-parmetis --download-metis)!')
    if self.metis.found:
      args.append('-DENABLE_METIS:BOOL=ON')
      args.append('-DMETIS_DIR:FILEPATH='+self.metis.directory)
    if self.zoltan.found or self.trilinos.found:
      args.append('-DENABLE_ZOLTAN:BOOL=ON')
      if self.zoltan.found:
        args.append('-DZOLTAN_DIR:FILEPATH='+self.zoltan.directory)
      else:
       args.append('-DZOLTAN_DIR:FILEPATH='+self.trilinos.directory)
      if self.parmetis.found:
        args.append('-DZOLTAN_NEEDS_ParMETIS=yes')
        args.append('-DParMETIS_DIR:FILEPATH='+self.parmetis.directory)
      if self.ptscotch.found:
        args.append('-DZOLTAN_NEEDS_PTSCOTCH=yes')
        args.append('-DPTSCOTCH_DIR:FILEPATH='+self.ptscotch.directory)

    if self.exodusii.found or self.trilinos.found:
      args.append('-DENABLE_EXODUSII:BOOL=ON')
      if self.exodusii.found:
        args.append('-DEXODUSII_DIR:FILEPATH='+self.exodusii.directory)
      else:
        args.append('-DEXODUSII_DIR:FILEPATH='+self.trilinos.directory)

    #  Need to pass -DMETIS_5 to C and C++ compiler flags otherwise assumes older Metis
    args = self.rmArgsStartsWith(args,['-DCMAKE_CXX_FLAGS:STRING','-DCMAKE_C_FLAGS:STRING'])
    args.append('-DCMAKE_C_FLAGS:STRING="'+self.updatePackageCFlags(self.getCompilerFlags())+' -DMETIS_5"')
    if hasattr(self.compilers, 'CXX'):
      self.pushLanguage('Cxx')
      args.append('-DCMAKE_CXX_FLAGS:STRING="'+self.updatePackageCxxFlags(self.getCompilerFlags())+' -DMETIS_5"')
    self.popLanguage()

    # mstk does not use the standard -DCMAKE_INSTALL_PREFIX
    args.append('-DINSTALL_DIR='+self.installDir)
    return args




import config.package
import os
import re

class Configure(config.package.Package):
  def __init__(self, framework):
    config.package.Package.__init__(self, framework)
    # NVSHMEM_MAJOR_VERSION and NVSHMEM_MINOR_VERSIO are not the NVSHMEM release version. They are the supported OpenSHMEM spec version.
    #self.versionname       = 'NVSHMEM_VENDOR_STRING' # the string has a format like "NVSHMEM v2.0.2"
    #self.versioninclude    = 'nvshmem_constants.h'
    #self.requiresversion   = 1
    self.buildLanguages    = ['CUDA'] # requires nvcc
    #self.functions         = ['nvshmem_init', 'nvshmem_finalize']
    self.includes          = ['nvshmem.h']
    self.liblist           = [['libnvshmem.a','libcuda.a'], ['nvshmem.lib','cuda.lib']]
    return

  def setupDependencies(self, framework):
    config.package.Package.setupDependencies(self, framework)
    self.cuda       = framework.require('config.packages.cuda',self)
    self.deps      = [self.cuda]
    return

  def versionToStandardForm(self,ver):
    '''Converts from "NVSHMEMv2.0.2" to standard notation 2.0.2'''
    return re.sub('[a-zA-Z ]*', '', ver)



'''A source code database

    SourceDB is a database of file information used to determine whether files
    should be rebuilt by the build system. All files names are stored relative
    to a given root, which is intended as the root of a Project.

    Relative or absolute pathnames may be used as keys, but absolute pathnames
    must fall under the database root. The value format is a tuple of the following:

      Checksum:     The md5 checksum of the file
      Mod Time:     The time the file was last modified
      Timestamp:    The time theentry was last modified
      Dependencies: A tuple of files upon which this entry depends

    This script also provides some default actions:

      - insert <database file> <filename>
        Inserts this file from the database, or updates its entry if it
        already exists.

      - remove <database file> <filename>
        Removes this file from the database. The filename may also be a
        regular expression.

'''
from __future__ import print_function
from __future__ import absolute_import
import logger

import errno
import os
import re
import time

import pickle
from hashlib import md5 as new_md5

class SourceDB (dict, logger.Logger):
  '''A SourceDB is a dictionary of file data used during the build process.'''
  includeRE = re.compile(r'^#include (<|")(?P<includeFile>.+)\1')
  isLoading = 0

  def __init__(self, root, filename = None):
    dict.__init__(self)
    logger.Logger.__init__(self)
    self.root       = root
    self.filename   = filename
    if self.filename is None:
      self.filename = os.path.join(str(root), 'bsSource.db')
    self.isDirty    = 0
    return

  def __str__(self):
    output = ''
    for source in self:
      (checksum, mtime, timestamp, dependencies) = self[source]
      output += source+'\n'
      output += '  Checksum:  '+str(checksum)+'\n'
      output += '  Mod Time:  '+str(mtime)+'\n'
      output += '  Timestamp: '+str(timestamp)+'\n'
      output += '  Deps:      '+str(dependencies)+'\n'
    return output

  def __setstate__(self, d):
    logger.Logger.__setstate__(self, d)
    # We have to prevent recursive calls to this when the pickled database is loaded in load()
    #   This is to ensure that fresh copies of the database are obtained after unpickling
    if not SourceDB.isLoading:
      SourceDB.isLoading = 1
      self.load()
      SourceDB.isLoading = 0
    return

  def getRelativePath(self, path):
    '''Returns a relative source file path using the root'''
    if os.path.isabs(path):
      root = str(self.root)
      if not path.startswith(root+os.sep):
        raise ValueError('Absolute path '+path+' conflicts with root '+root)
      else:
        path = path[len(root)+1:]
    return path

  def checkValue(self, value):
    '''Validate the value, raising ValueError for problems'''
    if not isinstance(value, tuple):
      raise ValueError('Source database values must be tuples, '+str(type(value))+' given')
    if not len(value) == 4:
      raise ValueError('Source database values must have 4 items, '+str(len(value))+' given')
    (checksum, mtime, timestamp, dependencies) = value
    if not isinstance(checksum, str):
      raise ValueError('Invalid checksum for source database, '+str(type(checksum))+' given')
    if not isinstance(mtime, int):
      raise ValueError('Invalid modification time for source database, '+str(type(mtime))+' given')
    elif mtime < 0:
      raise ValueError('Negative modification time for source database, '+str(mtime))
    if not isinstance(timestamp, float):
      raise ValueError('Invalid timestamp for source database, '+str(type(timestamp))+' given')
    elif timestamp < 0:
      raise ValueError('Negative timestamp for source database, '+str(timestamp))
    if not isinstance(dependencies, tuple):
      raise ValueError('Invalid dependencies for source database, '+str(type(dependencies))+' given')
    return value

  def __getitem__(self, key):
    '''Converts the key to a relative source file path using the root'''
    return dict.__getitem__(self, self.getRelativePath(key))

  def __setitem__(self, key, value):
    '''Converts the key to a relative source file path using the root, and checks the validity of the value'''
    self.isDirty = 1
    return dict.__setitem__(self, self.getRelativePath(key), self.checkValue(value))

  def __delitem__(self, key):
    '''Converts the key to a relative source file path using the root'''
    self.isDirty = 1
    return dict.__delitem__(self, self.getRelativePath(key))

  def __contains__(self, key):
    '''Converts the key to a relative source file path using the root'''
    return dict.__contains__(self, self.getRelativePath(key))

  def has_key(self, key):
    '''This method just calls self.__contains__(key)'''
    return self.__contains__(key)

  def items(self):
    '''Converts each key to a relative source file path using the root'''
    return [(self.getRelativePath(item[0]), item[1]) for item in dict.items(self)]

  def keys(self):
    '''Converts each key to a relative source file path using the root'''
    return map(self.getRelativePath, dict.keys(self))

  def update(self, d):
    '''Update the dictionary with the contents of d'''
    self.isDirty = 1
    for k in d:
      self[k] = d[k]
    return

  def getChecksum(source, chunkSize = 1024*1024):
    '''Return the md5 checksum for a given file, which may also be specified by its filename
       - The chunkSize argument specifies the size of blocks read from the file'''
    if hasattr(source, 'close'):
      f = source
    else:
      f = open(source)
    m = new_md5()
    size = chunkSize
    buf  = f.read(size)
    while buf:
      m.update(buf)
      buf = f.read(size)
    f.close()
    return m.hexdigest()
  getChecksum = staticmethod(getChecksum)

  def getModificationTime(source):
    t = os.path.getmtime(source)
    if isinstance(t, float):
      t = int(t)
    return t
  getModificationTime = staticmethod(getModificationTime)

  def updateSource(self, source, noChecksum = 0):
    self.isDirty = 1
    dependencies = ()
    try:
      (checksum, mtime, timestamp, dependencies) = self[source]
    except KeyError:
      pass
    self.logPrint('Updating '+source+' in source database', 3, 'sourceDB')
    if noChecksum:
      checksum   = ''
    else:
      checksum   = SourceDB.getChecksum(source)
    self[source] = (checksum, SourceDB.getModificationTime(source), time.time(), dependencies)
    return

  def clearSource(self, source):
    '''This removes source information, but preserved dependencies'''
    if source in self:
      self.isDirty = 1
      self.logPrint('Clearing '+source+' from source database', 3, 'sourceDB')
      (checksum, mtime, timestamp, dependencies) = self[source]
      self[source] = ('', 0, time.time(), dependencies)
    return

  def getDependencies(self, source):
    try:
      (checksum, mtime, timestamp, dependencies) = self[source]
    except KeyError:
      dependencies = ()
    return dependencies

  def addDependency(self, source, dependency):
    self.isDirty = 1
    dependencies = ()
    try:
      (checksum, mtime, timestamp, dependencies) = self[source]
    except KeyError:
      checksum = ''
      mtime    = 0
    if not dependency in dependencies:
      self.logPrint('Adding dependency '+dependency+' to source '+source+' in source database', 3, 'sourceDB')
      dependencies = dependencies+(dependency,)
    self[source] = (checksum, mtime, time.time(), dependencies)
    return

  def calculateDependencies(self):
    self.logPrint('Recalculating dependencies', 1, 'sourceDB')
    for source in self:
      self.logPrint('Calculating '+source, 3, 'sourceDB')
      (checksum, mtime, timestamp, dependencies) = self[source]
      newDep = []
      try:
        file = open(source)
      except IOError as e:
        if e.errno == errno.ENOENT:
          del self[source]
        else:
          raise e
      comps  = source.split('/')
      for line in file:
        m = self.includeRE.match(line)
        if m:
          filename  = m.group('includeFile')
          matchNum  = 0
          matchName = filename
          self.logPrint('  Includes '+filename, 3, 'sourceDB')
          for s in self:
            if s.find(filename) >= 0:
              self.logPrint('    Checking '+s, 3, 'sourceDB')
              c = s.split('/')
              for i in range(len(c)):
                if not comps[i] == c[i]: break
              if i > matchNum:
                self.logPrint('    Choosing '+s+'('+str(i)+')', 3, 'sourceDB')
                matchName = s
                matchNum  = i
          newDep.append(matchName)
      # Grep for #include, then put these files in a tuple, we can be recursive later in a fixpoint algorithm
      self[source] = (checksum, mtime, timestamp, tuple(newDep))
      file.close()

  def load(self):
    '''Load the source database from the saved filename'''
    filename = str(self.filename)
    if os.path.exists(filename):
      self.clear()
      self.logPrint('Loading source database from '+filename, 2, 'sourceDB')
      dbFile = open(filename)
      newDB  = pickle.load(dbFile)
      dbFile.close()
      self.update(newDB)
    else:
      self.logPrint('Could not load source database from '+filename, 1, 'sourceDB')
    return

  def save(self, force = 0):
    '''Save the source database to a file. The saved database with have path names relative to the root.'''
    if not self.isDirty and not force:
      self.logPrint('No need to save source database in '+str(self.filename), 2, 'sourceDB')
      return
    filename = str(self.filename)
    if os.path.exists(os.path.dirname(filename)):
      self.logPrint('Saving source database in '+filename, 2, 'sourceDB')
      dbFile = open(filename, 'w')
      pickle.dump(self, dbFile)
      dbFile.close()
      self.isDirty = 0
    else:
      self.logPrint('Could not save source database in '+filename, 1, 'sourceDB')
    return

class DependencyAnalyzer (logger.Logger):
  def __init__(self, sourceDB):
    logger.Logger.__init__(self)
    self.sourceDB  = sourceDB
    self.includeRE = re.compile(r'^#include (<|")(?P<includeFile>.+)\1')
    return

  def resolveDependency(self, source, dep):
    if dep in self.sourceDB: return dep
    # Choose the entry in sourceDB whose base matches dep,
    #   and who has the most path components in common with source
    # This should be replaced by an appeal to cpp
    matchNum   = 0
    matchName  = dep
    components = source.split(os.sep)
    self.logPrint('  Includes '+filename, 3, 'sourceDB')
    for s in self.sourceDB:
      if s.find(dep) >= 0:
        self.logPrint('    Checking '+s, 3, 'sourceDB')
        comp = s.split(os.sep)
        for i in range(len(comp)):
          if not components[i] == comp[i]: break
        if i > matchNum:
          self.logPrint('    Choosing '+s+'('+str(i)+')', 3, 'sourceDB')
          matchName = s
          matchNum  = i
    if not matchName in self.sourceDB: raise RuntimeError('Invalid #include '+matchName+' in '+source)
    return matchName

  def getNeighbors(self, source):
    file = open(source)
    adj  = []
    for line in file:
      match = self.includeRE.match(line)
      if match:
        adj.append(self.resolveDependency(source, m.group('includeFile')))
    file.close()
    return adj

  def calculateDependencies(self):
    '''Should this be a generator?
    First assemble the DAG using #include relations
    Then calculate the dependencies with all pairs shortest-path
      - I think Floyd-Warshell and N-source Dijkstra are just as good
    '''
    # Assembling DAG
    dag = {}
    for source in self.sourceDB:
      try:
        dag[source] = self.getNeighbors(self, source)
      except IOError as e:
        if e.errno == errno.ENOENT:
          del self[source]
        else:
          raise e
    # Finding all-pairs shortest path

if __name__ == '__main__':
  import sys
  try:
    if len(sys.argv) < 3:
      print('sourceDatabase.py <database filename> [insert | remove] <filename>')
    else:
      if os.path.exists(sys.argv[1]):
        dbFile   = open(sys.argv[1])
        sourceDB = pickle.load(dbFile)
        dbFile.close()
      else:
        sys.exit('Could not load source database from '+sys.argv[1])
      if sys.argv[2] == 'insert':
        if sys.argv[3] in sourceDB:
          self.logPrint('Updating '+sys.argv[3], 3, 'sourceDB')
        else:
          self.logPrint('Inserting '+sys.argv[3], 3, 'sourceDB')
        self.sourceDB.updateSource(sys.argv[3])
      elif sys.argv[2] == 'remove':
        if sys.argv[3] in sourceDB:
          sourceDB.logPrint('Removing '+sys.argv[3], 3, 'sourceDB')
          del self.sourceDB[sys.argv[3]]
        else:
          sourceDB.logPrint('Matching regular expression '+sys.argv[3]+' over source database', 1, 'sourceDB')
          removeRE = re.compile(sys.argv[3])
          removes  = list(filter(removeRE.match, sourceDB.keys()))
          for source in removes:
            self.logPrint('Removing '+source, 3, 'sourceDB')
            del self.sourceDB[source]
      else:
        sys.exit('Unknown source database action: '+sys.argv[2])
      sourceDB.save()
  except Exception as e:
    import traceback
    print(traceback.print_tb(sys.exc_info()[2]))
    sys.exit(str(e))
  sys.exit(0)



import config.package

# only install dmatgen from HPL, not the entire hpl package
class Configure(config.package.Package):
  def __init__(self, framework):
    config.package.GNUPackage.__init__(self, framework)
    self.version          = '2.3.0'
    self.maxversion       = '2.5.100000'
    self.versionname      = 'PACKAGE_VERSION'
    self.download         = ['https://www.netlib.org/benchmark/hpl/hpl-2.3.tar.gz']
    self.downloaddirnames = ['hpl']
    self.functions        = ['HPL_dmatgen']
    self.includes         = ['hpl.h']
    self.liblist          = [['libhpl.a']]
    self.complex          = 0
    self.precisions       = ['double']

  def setupDependencies(self, framework):
    config.package.Package.setupDependencies(self, framework)
    self.blasLapack = framework.require('config.packages.BlasLapack',self)
    self.mpi        = framework.require('config.packages.MPI',self)
    self.mathlib    = framework.require('config.packages.mathlib',self)
    self.deps       = [self.mpi,self.blasLapack,self.mathlib]

  def Install(self):
    import os, sys
    import config.base

    libDir         = os.path.join(self.installDir, 'lib')
    includeDir     = os.path.join(self.installDir, 'include')
    srcDir         = os.path.join(self.packageDir,'testing','matgen')
    makefile       = os.path.join(srcDir, 'makefile')

    g = open(makefile,'w')
    g.write('SHELL            = '+self.programs.SHELL+'\n')
    g.write('CP               = '+self.programs.cp+'\n')
    g.write('RM               = '+self.programs.RM+'\n')
    g.write('MKDIR            = '+self.programs.mkdir+'\n')
    g.write('OMAKE            = '+self.make.make+' '+self.make.noprintdirflag+'\n')

    g.write('CLINKER          = '+self.getLinker()+'\n')
    g.write('AR               = '+self.setCompilers.AR+'\n')
    g.write('ARFLAGS          = '+self.setCompilers.AR_FLAGS+'\n')
    g.write('AR_LIB_SUFFIX    = '+self.setCompilers.AR_LIB_SUFFIX+'\n')
    g.write('RANLIB           = '+self.setCompilers.RANLIB+'\n')
    g.write('SL_LINKER_SUFFIX = '+self.setCompilers.sharedLibraryExt+'\n')

    g.write('PREFIX           = '+self.installDir+'\n')
    g.write('LIBDIR           = '+libDir+'\n')
    g.write('INSTALL_LIB_DIR  = '+libDir+'\n')
    g.write('HPLLIB           = libhpl.$(AR_LIB_SUFFIX)\n')
    g.write('SHLIB            = libhpl\n')

    self.pushLanguage('C')
    cflags = self.updatePackageCFlags(self.getCompilerFlags())
    cflags += ' '+self.headers.toString('.')
    cflags += ' -fPIC'

    g.write('CC               = '+self.getCompiler()+'\n')
    g.write('CFLAGS           = '+cflags+'\n')
    self.popLanguage()
    g.write('clean:\n')
    g.write('	${RM} -f *.o\n')
    g.write('$(HPLLIB):\n')
    g.write('	$(CC) -I../../include $(CFLAGS)  -c *.c\n')
    g.write('	$(AR) $(ARFLAGS) $(HPLLIB) *.o\n')
    g.close()

    # compile & install
    if self.installNeeded(makefile):
      try:
        self.logPrintBox('Compiling HPL dmatgen; this may take several minutes')
        output1,err1,ret1  = config.package.Package.executeShellCommand('cd '+srcDir+' && make clean && make libhpl.'+self.setCompilers.AR_LIB_SUFFIX+' && make clean', timeout=250, log = self.log)
      except RuntimeError as e:
        raise RuntimeError('Error running make on HPL dmatgen: '+str(e))
      self.logPrintBox('Installing HPL dmatgen; this may take several minutes')
      output,err,ret = config.package.Package.executeShellCommand('mkdir -p '+os.path.join(self.installDir,'lib'), timeout=25, log=self.log)
      output,err,ret = config.package.Package.executeShellCommand('mkdir -p '+os.path.join(self.installDir,'include'), timeout=25, log=self.log)
      output2,err2,ret2  = config.package.Package.executeShellCommand('cp -f '+os.path.join(srcDir,'libhpl.'+self.setCompilers.AR_LIB_SUFFIX)+' '+os.path.join(self.installDir,'lib'), timeout=60, log = self.log)
      output2,err2,ret2  = config.package.Package.executeShellCommand('cp -f '+os.path.join(self.packageDir, 'include', '*.h')+' '+includeDir, timeout=60, log = self.log)
      self.postInstall(output1+err1+output2+err2,makefile)
    return self.installDir

  def configureLibrary(self):
    config.package.Package.configureLibrary(self)
    return
